global:
    cpp_namespace: "mongo"

imports:
    - "mongo/idl/basic_types.idl"

structs:
    SorterRange:
        description: "The range of data that was sorted and spilled to disk."
        strict: true
        fields:
            startOffset:
                description: "Tracks where in the file we started writing this data range."
                type: long
            endOffset:
                description: "Tracks where in the file we finished writing this data range."
                type: long
            checksum:
                description: "Tracks the hash of all data objects spilled to disk."
                type: long
                validator: { gte: 0 }
