# Copyright (C) 2020-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#
global:
    cpp_namespace: "mongo"

imports:
    - "mongo/idl/basic_types.idl"

enums:
    MessageSeverity:
        description: "Severity level to log message at"
        type: string
        values:
            kSevere: "severe"
            kError: "error"
            kWarning: "warning"
            kInfo: "info"
            kLog: "log"
            kDebug: "debug"

commands:
    logMessage:
        description: "Log a message on the server"
        command_name: logMessage
        namespace: type
        api_version: ""
        type: string
        cpp_name: LogMessageCommand
        strict: true
        fields:
            extra:
                description: "Structured data to include with log message"
                type: object
                optional: true
            severity:
                description: "Severity of log message"
                type: MessageSeverity
                default: kLog
            debugLevel:
                description: "When using 'debug' severity, what level of debug is to be used?"
                type: int
                optional: true
                validator:
                    gte: 1
                    lte: 5

    ping:
        description: "Parser for the 'ping' command."
        command_name: ping
        strict: true
        api_version: 1
        access_check:
            none: true
        namespace: ignored
        reply_type: OkReply
