# Copyright (C) 2019-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#

# This IDL file describes the BSON format for a LogicalSessionId, and
# handles the serialization to and deserialization from its BSON representation
# for that class.

global:
  cpp_namespace: "mongo"

imports:
  - "mongo/idl/basic_types.idl"

structs:

  CommitTypeStats:
    description: "Contains metrics for a type of commit"
    fields:
      initiated:
        type: long
        default: 0
      successful:
        type: long
        default: 0
      successfulDurationMicros:
        type: long
        default: 0

  CommitTypes:
    description: "Contains metrics for each type of commit"
    fields:
      noShards:
        type: CommitTypeStats
      singleShard:
        type: CommitTypeStats
      singleWriteShard:
        type: CommitTypeStats
      readOnly:
        type: CommitTypeStats
      twoPhaseCommit:
        type: CommitTypeStats
      recoverWithToken:
        type: CommitTypeStats

  RouterTransactionsStats:
    description: "A struct representing the section of the server status
                  command on mongos with information about sharded transactions"
    strict: true
    fields:
      currentOpen:
        type: long
        default: 0
      currentActive:
        type: long
        default: 0
      currentInactive:
        type: long
        default: 0
      totalStarted:
        type: long
        default: 0
      totalCommitted:
        type: long
        default: 0
      totalAborted:
        type: long
        default: 0
      abortCause:
        type: object
      totalContactedParticipants:
        type: long
        default: 0
      totalParticipantsAtCommit:
        type: long
        default: 0
      totalRequestsTargeted:
        type: long
        default: 0
      commitTypes:
        type: CommitTypes
