/**
 *    Copyright (C) 2018-present MongoDB, Inc.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the Server Side Public License, version 1,
 *    as published by MongoDB, Inc.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    Server Side Public License for more details.
 *
 *    You should have received a copy of the Server Side Public License
 *    along with this program. If not, see
 *    <http://www.mongodb.com/licensing/server-side-public-license>.
 *
 *    As a special exception, the copyright holders give permission to link the
 *    code of portions of this program with the OpenSSL library under certain
 *    conditions as described in each individual source file and distribute
 *    linked combinations including the program with the OpenSSL library. You
 *    must comply with the Server Side Public License in all respects for
 *    all of the code used other than as permitted herein. If you modify file(s)
 *    with this exception, you may extend this exception to your version of the
 *    file(s), but you are not obligated to do so. If you do not wish to do so,
 *    delete this exception statement from your version. If you delete this
 *    exception statement from all source files in the program, then also delete
 *    it in the license file.
 */

#pragma once

#include <memory>

#include "mongo/base/string_data.h"
#include "mongo/db/update/modifier_node.h"

namespace mongo {

/**
 * Represents the application of $inc or $mul to the value at the end of a path.
 */
class ArithmeticNode : public ModifierNode {
public:
    enum class ArithmeticOp { kAdd, kMultiply };

    explicit ArithmeticNode(ArithmeticOp op) : _op(op) {}

    Status init(BSONElement modExpr, const boost::intrusive_ptr<ExpressionContext>& expCtx) final;

    std::unique_ptr<UpdateNode> clone() const final {
        return std::make_unique<ArithmeticNode>(*this);
    }

    void setCollator(const CollatorInterface* collator) final {}

    void acceptVisitor(UpdateNodeVisitor* visitor) final {
        visitor->visit(this);
    }

protected:
    ModifyResult updateExistingElement(mutablebson::Element* element,
                                       const FieldRef& elementPath) const final;
    void setValueForNewElement(mutablebson::Element* element) const final;

    bool allowCreation() const final {
        return true;
    }

private:
    StringData operatorName() const final {
        switch (_op) {
            case ArithmeticNode::ArithmeticOp::kAdd:
                return "$inc";
            case ArithmeticNode::ArithmeticOp::kMultiply:
                return "$mul";
        }
        MONGO_UNREACHABLE;
    }

    BSONObj operatorValue() const final {
        return BSON("" << _val);
    }

    ArithmeticOp _op;
    BSONElement _val;
};

}  // namespace mongo
