# Copyright (C) 2019-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#
global:
    cpp_namespace: "mongo"

imports:
    - "mongo/idl/basic_types.idl"
    - "mongo/db/repl/read_concern_args.idl"
    - "mongo/db/write_concern_options.idl"

enums:
    DefaultWriteConcernSource:
        description: "The source of the default write concern"
        type: string
        values:
            # The default write concern was set implicitly by the server
            kImplicit: "implicit"
            # The default write concern was set globally through setDefaultRWConcern
            kGlobal: "global"
    DefaultReadConcernSource:
        description: "The source of the default read concern"
        type: string
        values:
            # The default read concern was set implicitly by the server.
            kImplicit: "implicit"
            # The default read concern was set globally through setDefaultRWConcern.
            kGlobal: "global"

structs:
    RWConcernDefault:
        description: "Represents a set of read/write concern defaults, and associated metadata"
        strict: false
        fields:
            defaultReadConcern:
                description: "The default read concern"
                type: ReadConcern
                optional: true
            defaultWriteConcern:
                description: "The default write concern"
                type: WriteConcern
                optional: true
            updateOpTime:
                description: "The optime of when the default read or write concern was last set. On
                              replica sets it advances with the primary's optime and on clusters it
                              advances with the config server primary's optime."
                type: timestamp
                optional: true
            updateWallClockTime:
                description: "The wallclock time when the default read or write concern was last set
                              by an administrator. This value is only informational and must not be
                              used for any recency comparisons."
                type: date
                optional: true
            defaultWriteConcernSource:
                description: "The source of the default write concern."
                type: DefaultWriteConcernSource
                optional: true
            defaultReadConcernSource:
                description: "The source of the default read concern"
                type: DefaultReadConcernSource
                optional: true
