# Copyright (C) 2021-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#

# cleanupReshardCollection IDL file

global:
    cpp_namespace: "mongo"

imports:
    - "mongo/idl/basic_types.idl"

commands:
  cleanupReshardCollection:
    description: "The public cleanupReshardCollection command on mongos."
    command_name: cleanupReshardCollection
    strict: false
    namespace: type
    api_version: ""
    type: namespacestring

  _configsvrCleanupReshardCollection:
    command_name: _configsvrCleanupReshardCollection
    cpp_name: ConfigsvrCleanupReshardCollection
    description: "The internal cleanupReshardCollection command on the config server."
    strict: false
    namespace: type
    api_version: ""
    type: namespacestring

  _shardsvrCleanupReshardCollection:
    command_name: _shardsvrCleanupReshardCollection
    cpp_name: ShardsvrCleanupReshardCollection
    description: "The internal cleanupReshardCollection command on the shard server."
    strict: false
    namespace: type
    api_version: ""
    type: namespacestring
    fields:
      reshardingUUID:
        type: uuid
        description: "Identifier for a resharding operation."
