
# Copyright (C) 2019-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#

global:
    cpp_namespace: "mongo"
    configs:
        section: 'Windows Service Control Manager options'
        source: [ yaml, cli, ini ]

configs:
    install:
        description: 'Install Windows service'
        arg_vartype: Switch
        source: [ cli, ini ] 
    remove:
        description: 'Remove Windows service'
        arg_vartype: Switch
        source: [ cli, ini ]
    reinstall:
        description: 'Reinstall Windows service (equivalent to --remove followed by --install)'
        arg_vartype: Switch
        source: [ cli, ini ]
    service:
        description: 'Start mongodb service'
        arg_vartype: Switch
        source: [ cli, ini ]
        hidden: true

    'processManagement.windowsService.serviceName':
        description: 'Windows service name'
        short_name: serviceName
        arg_vartype: String
    'processManagement.windowsService.displayName':
        description: 'Windows service display name'
        short_name: serviceDisplayName
        arg_vartype: String
    'processManagement.windowsService.description':
        description: 'Windows service description'
        short_name: serviceDescription
        arg_vartype: String
    'processManagement.windowsService.serviceUser':
        description: 'Account for service execution'
        short_name: serviceUser
        arg_vartype: String
    'processManagement.windowsService.servicePassword':
        description: 'Password used to authenticate serviceUser'
        short_name: servicePassword
        arg_vartype: String
        redact: true

