# Copyright (C) 2020-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#

global:
  cpp_namespace: "mongo"
  cpp_includes:
    - "mongo/client/read_preference.h"
    - "mongo/db/repl/tenant_migration_util.h"

imports:
  - "mongo/client/read_preference_setting.idl"
  - "mongo/db/repl/tenant_migration_state_machine.idl"
  - "mongo/idl/basic_types.idl"
  - "mongo/s/sharding_types.idl"

structs:
    DonorStartMigrationResponse:
        description: "Response of the donorStartMigration command"
        strict: false
        fields:
            state:
                type: TenantMigrationDonorState
                description: "The state of the tenant migration."
            abortReason:
                type: object
                description: "The error that caused the migration to abort."
                optional: true

commands:
  donorStartMigration:
    description: "Parser for the 'donorStartMigration' command."
    command_name: donorStartMigration
    strict: true
    namespace: ignored
    api_version: ""
    fields:
      migrationId:
        description: "Unique identifier for the tenant migration."
        type: uuid
      recipientConnectionString:
        description: "The URI string that the donor will utilize to create a connection with the recipient."
        type: string
        validator:
          callback: "tenant_migration_util::validateConnectionString"
      tenantId:
        description: "The prefix from which the migrating database will be matched. The prefixes 'admin', 'local', 'config', the empty string, are not allowed."
        type: string
        validator:
          callback: "tenant_migration_util::validateDatabasePrefix"
      readPreference:
        description: "The read preference settings that the donor will pass on to the recipient."
        type: readPreference
      donorCertificateForRecipient:
        description: >-
          The SSL certificate and private key that the donor should use to authenticate to the
          recipient.
        type: TenantMigrationPEMPayload
        # TODO (SERVER-54085): Remove server parameter tenantMigrationDisableX509Auth.
        optional: true
      recipientCertificateForDonor:
        description: >-
          The SSL certificate and private key that the recipient should use to authenticate to the
          donor.
        type: TenantMigrationPEMPayload
        # TODO (SERVER-54085): Remove server parameter tenantMigrationDisableX509Auth.
        optional: true

  donorForgetMigration:
    description: "Parser for the 'donorForgetMigration' command."
    command_name: donorForgetMigration
    strict: true
    namespace: ignored
    api_version: ""
    fields:
      migrationId:
        description: "Unique identifier for the tenant migration."
        type: uuid

  donorAbortMigration:
    description: "Parser for the 'donorAbortMigration' command."
    command_name: donorAbortMigration
    strict: true
    namespace: ignored
    api_version: ""
    fields:
      migrationId:
        description: "Unique identifier for the tenant migration."
        type: uuid
