# Copyright (C) 2020-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#

# This file defines the document used for storing progress by the resharding oplog applier.

global:
    cpp_namespace: "mongo"

imports:
    - "mongo/idl/basic_types.idl"
    - "mongo/db/s/resharding/donor_oplog_id.idl"
    - "mongo/s/resharding/common_types.idl"
    - "mongo/s/sharding_types.idl"

structs:
    ReshardingOplogApplierProgress:
        description: "Used for storing the progress made by the resharding oplog applier."
        # Use strict:false to avoid complications around upgrade/downgrade. This isn't technically
        # required for resharding because durable state from all resharding operations is cleaned up
        # before the upgrade or downgrade can complete.
        strict: false
        fields:
            _id:
                type: ReshardingSourceId
                description: "The identifier for the source of the oplog entries."
                cpp_name: oplogSourceId
            progress:
                type: ReshardingDonorOplogId
                description: >-
                    The minimum point where the resharding oplog applier can start without missing
                    any oplog it needs to process.
            numEntriesApplied:
                type: long
                description: >-
                    The total number of oplog entries that have been applied by this
                    ReshardingOplogApplier.
