# Copyright (C) 2018-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http:#www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#

global:
    cpp_namespace: "mongo"

imports:
    - "mongo/idl/basic_types.idl"
    - "mongo/db/query/explain_verbosity.idl"

commands:
    explain:
        description: "Parser for the explain command"
        command_name: explain
        cpp_name: ExplainCommandRequest
        namespace: type
        type: object
        strict: true
        api_version: "1"
        access_check:
            ignore: true
        fields:
            verbosity:
                description: "The verbosity for explain command."
                type: Verbosity
                default : kExecAllPlans
            # TODO SERVER-48560: we ingest these fields for compatibility with 4.4,
            # whose mongoS incorrectly adds them to the explain command for an
            # aggregation instead of adding them into the wrapped aggregate command
            # itself. Remove these fields when 5.0 becomes last-lts.
            collation:
                type: IDLAnyType
                ignore: true
                unstable: true
            use44SortKeys:
                type: IDLAnyType
                ignore: true
                unstable: true
            useNewUpsert:
                type: IDLAnyType
                ignore: true
                unstable: true
        # Dummy reply type as we won't use it to parse explain reply.
        reply_type: OkReply