# Copyright (C) 2018-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#
global:
    cpp_namespace: "mongo"


imports:
    - "mongo/db/auth/access_checks.idl"
    - "mongo/db/auth/action_type.idl"
    - "mongo/idl/basic_types.idl"

structs:
    ListDatabasesReplyItem:
        description: "Individual result"
        fields:
            name: string
            sizeOnDisk:
                type: long
                optional: true
            empty:
                type: bool
                optional: true
            shards:
                type: object_owned
                optional: true

    ListDatabasesReply:
        description: "The listDatabases command's reply."
        fields:
            databases: array<ListDatabasesReplyItem>
            totalSize:
                type: long
                optional: true
            totalSizeMb:
                type: long
                optional: true

commands:
    listDatabases:
        description: "listDatabases Command"
        command_name: "listDatabases"
        cpp_name: ListDatabasesCommand
        namespace: ignored
        api_version: "1"
        access_check:
            complex:
                - check: is_authorized_for_any_action_on_any_resource_in_db
                - privilege:
                    resource_pattern: cluster
                    action_type: listDatabases
        strict: true
        fields:
            nameOnly:
                description: "Return just the database name without metadata"
                type: safeBool
                default: false
            authorizedDatabases:
                description: "Only return those databases for which the user is authorized"
                type: bool
                optional: true
            filter:
                description: "Filter description to limit results"
                type: object
                optional: true
        reply_type: ListDatabasesReply
