# Copyright (C) 2021-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#
global:
    cpp_namespace: "mongo"

enums:
    AccessCheck:
        description: "List of supported access checks in AuthorizationSession"
        type: string
        values:
            kCheckAuthorizedToListCollections : "check_authorized_to_list_collections"
            kCheckCursorSessionPrivilege : "check_cursor_session_privilege"
            kClearImpersonatedUserData : "clear_impersonated_user_data"
            kGetAuthenticatedRoleNames : "get_authenticated_role_names"
            kGetAuthenticatedUserNames : "get_authenticated_user_names"
            kGetImpersonatedRoleNames : "get_impersonated_role_names"
            kGetImpersonatedUserNames : "get_impersonated_user_names"
            kGetSingleUser : "get_single_user"
            kIsAuthenticated :  "is_authenticated"
            kIsAuthenticatedAsUserWithRole : "is_authenticated_as_user_with_role"
            kIsAuthorizedForAnyActionOnAnyResourceInDB : "is_authorized_for_any_action_on_any_resource_in_db"
            kIsAuthorizedForAnyActionOnResource : "is_authorized_for_any_action_on_resource"
            kIsAuthorizedToChangeAsUser : "is_authorized_to_change_as_user"
            kIsAuthorizedToCreateRole : "is_authorized_to_create_role"
            kIsAuthorizedToParseNamespaceElement : "is_authorized_to_parse_namespace_element"
            kIsCoAuthorized :  "is_coauthorized"
            kIsCoauthorizedWith : "is_coauthorized_with"
            kIsCoauthorizedWithClient : "is_coauthorized_with_client"
            kIsImpersonating : "is_impersonating"
            kLookupUser :  "lookup_user"
            kShouldIgnoreAuthChecks :  "should_ignore_auth_checks"
            kIsUsingLocalhostBypass : "is_using_localhost_bypass" # Called in common code in commands.cpp dispatch
