# Copyright (C) 2018-present MongoDB, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Server Side Public License, version 1,
# as published by MongoDB, Inc.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Server Side Public License for more details.
#
# You should have received a copy of the Server Side Public License
# along with this program. If not, see
# <http://www.mongodb.com/licensing/server-side-public-license>.
#
# As a special exception, the copyright holders give permission to link the
# code of portions of this program with the OpenSSL library under certain
# conditions as described in each individual source file and distribute
# linked combinations including the program with the OpenSSL library. You
# must comply with the Server Side Public License in all respects for
# all of the code used other than as permitted herein. If you modify file(s)
# with this exception, you may extend this exception to your version of the
# file(s), but you are not obligated to do so. If you do not wish to do so,
# delete this exception statement from your version. If you delete this
# exception statement from all source files in the program, then also delete
# it in the license file.
#

global:
  cpp_namespace: "mongo"

imports:
  - "mongo/idl/basic_types.idl"

types:
  _id_key:
    bson_serialization_type: any
    description: "An _id value."
    cpp_type: "mongo::BSONKey"
    serializer: "mongo::BSONKey::serializeToBSON"
    deserializer: "mongo::BSONKey::parseFromBSON"

enums:
  Severity:
    description: "The severity of a healthlog entry."
    type: string
    values:
      Info: "info"
      Warning: "warning"
      Error: "error"
  Scope:
    description: "The scope covered by a healthlog entry."
    type: string
    values:
      Cluster: "cluster"
      Node: "node"
      Database: "database"
      Collection: "collection"
      Index: "index"
      Document: "document"

structs:
  HealthLogEntry:
    description: "An entry in system.local.healthlog."
    fields:
      namespace:
          cpp_name: nss
          type: namespacestring
      timestamp: date
      severity: Severity
      msg: string
      scope: Scope
      operation: string
      data: object
